/*
    Copyright 2008 Piotr Domagala and Jakub Mielczarek
    This file is part of the Filtertool.

    Filtertool is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

/**
  * @author Piotr Domagala
  * @author Jakub Mielczarek
  */
#include <vector>
#include <string>
#include <iostream>
#include <cstdlib>
#include <sstream>
#include <fstream>
#include <ctime>
#include "DataHandler.h" 
using namespace std;
/* Filtr działa w następujący sposób :
    dla opcji -f 1,gt12,0,1,2
    oznacza to: wypisz wiersze w kolumnach 0,1,2 dla których wartości w pierwszej kolumnie jest większa od 12
*/ 
/*! \brief DataOperations: class to handle data and perform operations on them
  This class inherits DataHandler (interface). This construction allows
  to dynamicly select the temlate argument.
  \sa DataHandler
 */
template <typename T> 
class DataOperations: public DataHandler
{
public:
 /*! This is the class constructor. 
   It takes no arguments and basically it is doing nothing
   except setting isJoined to false;
   \sa ~DataOperations
  */
  DataOperations();
 /*! This is the class destructor. 
   It takes no arguments and clears all vectors
   \sa DataOperations
  */
  ~DataOperations();
private:
  size_t col_cond;
  bool isJoined;
  char c;
  vector<string> file_names;
  vector<vector<vector<T> > > all_values;
  vector<vector<string> > all_comments;
  vector<vector<string> > results;
  vector<string> file_out;
  vector<vector<T> > tab_tmp;
  vector<T> val;
  vector<T> vall_tmp;
  vector<size_t> col_out;
  vector<double> conditions;
  vector<string> kind_of_cond;
  vector<size_t> out;
  string command;
  DataOperations(DataOperations &) {} // Private copy constructor prevents from creating two same objects
public:
 /*! This function swaps two columns in selected file.
   \param command is a command line argument of -s (--swap) option
  */
  virtual void swap(string command);
 /*! Function to collect input file names.
   \param names is a list of files separated by ','
  */
  virtual void parseFiles(string names ); 
 /*! This function loads data to memory.
  If input files have been set, this function will open each of them
  and read content. Otherwise it will start reading standard input.
  The EOF symbol ends reading.
  */
  virtual void load();
 /*! This function joins all input to one output.
    Actually it works only when two or more files were specified on the input
 */
  virtual void join();
 /*! Funtion to calculate mean.
   It can works on columns or rows.
   \param command is a value of -m (--mean) option
  */
  virtual void mean(char arg);
 /*! This function is called on the end, when we want to prsize_t all.
  Function will decide where to save the data one file, many files
  or prsize_t to standard output.
  */
  virtual void print();
 /*! Setter for command that started the program.
   */
  virtual void setCommand(string command);
 /*! This function parses the filtr option.
   \param command is a command line argument of -f (--filtr) option.
  */
  virtual void setFilter(string arg);
/*! Sets the output file names.
  \param names is a list of output files specified by -o (--output-files) option and separated by ','.
*/
  virtual void setOutputFiles(string names);
protected:
/*! Protected function to prsize_t data on standard output */
  void printToStd();
/*! Protected function to prsize_t data to one file */
  void printToFile();
/*! Protected function to prsize_t data to many files */
  void printToFiles();
/*! This function starts filtering data with condidions given in setFiltr 
  \sa setFiltr
*/
  size_t startFiltering();
  
};

template <typename T>
DataOperations<T>::DataOperations(): isJoined(false)
{

}
template <typename T>
DataOperations<T>::~DataOperations()
{
  all_comments.clear();
  all_values.clear();
  file_names.clear();
  file_out.clear();
  conditions.clear();
  kind_of_cond.clear();
  results.clear();
  tab_tmp.clear();
  out.clear();
  kind_of_cond.clear();
  vall_tmp.clear();
  col_out.clear();
}
template <typename T>
void DataOperations<T>::parseFiles(string names)
{
  char delim = ',';
  string name;
  istringstream iss (names);
  while (getline(iss, name, delim))
  {
    file_names.push_back(name);
  }
  load();
}
template <typename T>
void DataOperations<T>::join()
{
  isJoined=true;
}
template <typename T>
void DataOperations<T>::load()
{
  vector<vector<T> > tab, tab_tmp;
  vector<string> comments, com_tmp;
  vector<T> values, val_tmp;
  string s, line;
  char buf[100];
  T liczba;
  size_t found, found1;
  const char* white=" \t\r\n";
  size_t ile=file_names.size();
  if(file_names.size()==0) ile=1;
  for(size_t i=0;i<ile;i++)
  { 
    istream *in;
    ifstream infile;
    if (file_names.size()!=0)
    {
      infile.open(file_names[i].c_str());
      in = &infile;
    }
    else
    {
      in = &cin;
    }
    while(in->getline(buf, 100))
    {
      line = string(buf);
      found=line.find_first_not_of(white);
      if (found!=string::npos && line[found]=='#')
      com_tmp.push_back(line.substr(found));
      else if(found!=string::npos && line[found]!='#')
      {
	line=line+"\r\n";
	do
	{
	  found1=line.find_first_of(white, found);
	  if(found1!=string::npos) 
	  s=line.substr(found, found1-found);
	  found=line.find_first_not_of(white, found1);
	  liczba=(T)atof(s.c_str());
	  if(liczba!=0 || (liczba==0 && (s=="0" || s=="0.0"))) 
	  val_tmp.push_back(liczba);
	}while(found!=string::npos && found1!=string::npos);
	if(!val_tmp.empty())
	{
	  values=val_tmp;
	  tab_tmp.push_back(values);
	  val_tmp.clear();
	}
      }
    }
    if(!tab_tmp.empty())
    {
      tab=tab_tmp;
      all_values.push_back(tab);
      tab_tmp.clear();
    }
    if(!com_tmp.empty())
    {
      comments=com_tmp;
      all_comments.push_back(comments);
      com_tmp.clear();
    }
  }
  tab_tmp.clear();
  com_tmp.clear();
  val_tmp.clear();
  comments.clear();
  values.clear();
  tab.clear();
}
template <typename T>
void DataOperations<T>::swap(string arg)
{
  size_t n, m, file;
  n=atoi(arg.c_str());
  size_t found=arg.find_first_of(',');
  size_t found2=arg.find_last_of(',');
  m=atoi(arg.substr(found+1, found2).c_str());
  file = atoi(arg.substr(found2+1).c_str());
  size_t begin;
  size_t end;
  T tmp;
  n--; m--; file--;
  if(isJoined)
  {
    begin = 0;
    end = all_values.size();
  }
  else
  {
    begin = file;
    end = file+1;
  }
  for(size_t i= begin; i<end; i++) 
    if(n>=all_values[i][0].size() || m>=all_values[i][0].size() || n==m || file > all_values.size() || begin>=all_values.size())
    {
      cout<<"wrong swap argument!"<<endl;
      exit(-1);
    }
  for( /* begin  */; begin<end; begin++)
    for(size_t j=0; j<all_values[begin].size();j++)
    {
	tmp=all_values[begin][j][n];
	all_values[begin][j][n]=all_values[begin][j][m];
	all_values[begin][j][m]=tmp;
    }
      return;
}
template <typename T>
void DataOperations<T>::mean(char arg)
{
  T tmp=0;
  int l=0;
  vector<T> tmp_vec;
  switch(arg)
  {
    case 'l':
      for(size_t i=0;i<all_values.size();i++)
      {
	for(size_t j=0;j<all_values[i].size();j++)
	{
	  for(size_t k=0;k<all_values[i][j].size();k++)
	  {
	    tmp+=all_values[i][j][k];
	    l++;
	  }
	  tmp=tmp/ ((float)l);
	  all_values[i][j].push_back(tmp);
	  tmp =0 ;
	  l = 0;
	}
	tmp_vec.clear();
      }
      break;
    case 'c':
      for(size_t i=0;i<all_values.size();i++)
      {
	tmp_vec.clear();
	tmp_vec.resize(all_values[i][0].size());
	for(size_t j=0;j<all_values[i].size();j++)
	{
	  for(size_t k=0;k<all_values[i][j].size();k++)
	    tmp_vec[k]+=all_values[i][j][k];
	}
	for(size_t k=0; k<all_values[i][0].size();k++)
	  tmp_vec[k] = tmp_vec[k] / (float)all_values[i].size();
	all_values[i].push_back(tmp_vec);
      }
      break;
    default:
      cout<<"Wrong argument in mean"<<endl;
      return;
      break;
  }
}
template <typename T>
void DataOperations<T>::print()
{
  if(file_out.size()==0)
    printToStd();
  else if(file_out.size()==1 && (isJoined || file_names.size()==1))
    printToFile();
  else if(file_names.size() == file_out.size())
    printToFiles();
  else
  {
    int c=1;
    while(c)
    {
      cout<<"\nWARNING!!!Output and input file lists doesnt match to any standard pattern \nYou have three options:\n1. Print all to standard output.\n2. Print all to first file on the output file list.\n\n0. Exit without saving.\n(select 1, 2 or 0): ";
      cin>>c;
      switch (c)
      {
	case 1:
	  printToStd();
	  c=0;
	  break;
	case 2:
	  printToFile();
	  c=0;
	  break;
	case 0:
	  break;
	default:
	  break;
      }
    }
  }
}
template <typename T>
void DataOperations<T>::printToStd()
{
  time_t rawtime;
  struct tm * timeinfo;
  time ( &rawtime );
  timeinfo = localtime ( &rawtime );
  cout<<"#"<<command<<endl;
  cout<<"#"<<asctime (timeinfo)<<endl;
  for(size_t i=0;i<all_values.size();i++)
    for(size_t j=0;j<all_values[i].size(); j++)
    {
      for(size_t k=0;k<all_values[i][j].size();k++)
	cout<<all_values[i][j][k]<<"\t";
      cout<<endl;
    }
}
template <typename T>
void DataOperations<T>::printToFile()
{
  time_t rawtime;
  struct tm * timeinfo;
  time ( &rawtime );
  timeinfo = localtime ( &rawtime );
  ofstream outfile;
  outfile.open(file_out[0].c_str());
  outfile<<"#"<<command<<endl;
  outfile<<"#"<<asctime (timeinfo)<<endl;
  for(size_t i=0;i<all_values.size();i++)
    for(size_t j=0;j<all_values[i].size(); j++)
    {
      for(size_t k=0;k<all_values[i][j].size();k++)
	outfile<<all_values[i][j][k]<<"\t";
      outfile<<endl;
    }
}
template <typename T>
void DataOperations<T>::printToFiles()
{
  time_t rawtime;
  struct tm * timeinfo;
  time ( &rawtime );
  timeinfo = localtime ( &rawtime );
  ofstream outfile;
  for(size_t i=0;i<all_values.size();i++)
  {
    outfile.open(file_out[i].c_str());
    outfile<<"#"<<command<<endl;
    outfile<<"#"<<asctime (timeinfo)<<endl;
    for(size_t j=0;j<all_values[i].size(); j++)
    {
      for(size_t k=0;k<all_values[i][j].size();k++)
	outfile<<all_values[i][j][k]<<"\t";
      outfile<<endl;
    }
    outfile.close();
  }
}
template <typename T>
void DataOperations<T>::setCommand(string cmd)
{
  command = cmd;
}
template <typename T>
void DataOperations<T>::setFilter(string arg)
{
  int z,outt;
  float val1,val2;
  char *war1=new char[2];
  char *war2=new char[2];
  char *tmp=new char[100];
  char *tmp2=new char[100];
  c='0';
  int aa;
  const char *b=arg.c_str();
  z=sscanf(b,"%d,%2s%f,%s",&aa,war1,&val1,tmp);
  col_cond = aa;
  if(z<3 || (strcmp(war1,"lt") && strcmp(war1,"gt") && strcmp(war1,"le") && strcmp(war1,"ge") && strcmp(war1,"eq") && strcmp(war1,"ne"))) {cout<<"Wrong filtr argument"<<endl;exit(1);}
  col_cond--;
  kind_of_cond.push_back(string(war1));
  conditions.push_back(val1);
  z=sscanf(tmp, "%c",&c);
  if(c!='0')
    if(c=='o' || c=='a')
    {
	strcpy(tmp2,tmp);
	strcpy(tmp,"");
	z=sscanf(tmp2,"%*c,%2s%f,%s",war2,&val2,tmp);
	if((z<2 && !strcmp(tmp,"")) || (strcmp(war2,"lt") && strcmp(war2,"gt") && strcmp(war2,"le") && strcmp(war2,"ge") && strcmp(war2,"eq") && strcmp(war1,"ne")))
	{cout<<"Wrong filtr argument"<<endl;exit(1);}
	kind_of_cond.push_back(string(war2));
	conditions.push_back(val2);
    }
    else {cout<<"Wrong filtr argument"<<endl;exit(1);}
  if(strcmp(tmp,""))
    while(strcmp(tmp,""))
    {
	strcpy(tmp2,tmp);
	strcpy(tmp,"");
	z=sscanf(tmp2, "%d,%s",&outt,tmp);
	if((z<1 && !strcmp(tmp,"")))
	{cout<<"Wrong filtr argument"<<endl;exit(1);}
	out.push_back(--outt);
    }
  startFiltering();
}
template <typename T>
size_t DataOperations<T>::startFiltering()
{
	size_t a=0;
	cout<<endl;
	for(size_t i=0; i<all_values.size(); i++)
		for(size_t j=0; j<all_values[i].size();j++)
		{
			for(size_t k=0;k<kind_of_cond.size();k++)
			{
				if(all_values[i][j].size()<=col_cond) {a=0;break;}
				if(kind_of_cond[k].compare("eq")==0)
						if(all_values[i][j][col_cond]==conditions[k])
							a++;
						else {a=0;}
				else if(kind_of_cond[k].compare("lt")==0)
						if(all_values[i][j][col_cond]<conditions[k])
							a++;
						else {a=0;}
				else if(kind_of_cond[k].compare("le")==0)
						if(all_values[i][j][col_cond]<=conditions[k])
							a++;
						else {a=0;}
				else if(kind_of_cond[k].compare("gt")==0)
						if(all_values[i][j][col_cond]>conditions[k])
							a++;
						else {a=0;}
				else if(kind_of_cond[k].compare("ge")==0)
						if(all_values[i][j][col_cond]>=conditions[k])
							a++;
						else {a=0;}
				else if(kind_of_cond[k].compare("ne")==0)
						if(all_values[i][j][col_cond]!=conditions[k])
							a++;
						else {a=0;}
				if(c=='o') {if(a!=0) break;}//gdy or jeśli spełniony pierwszy warunek przerywamy dalsze sprawdzanie i wypisujemy
				else if(c=='a') {if(a==0) break;}//gdy and jeśli nie spełniony pierwszy warunek przerywamy dlasze sprawdzanie i nie wypisujemy
			}
			if(a!=0)
			{
				if(out.size()>0)
					for(size_t l=0;l<out.size();l++){
						if(all_values[i][j].size()<=out[l]); //do nothing
						else
							vall_tmp.push_back(all_values[i][j][out[l]]);
					}
				else
					for(size_t l=0;l<all_values[i][j].size();l++){
							vall_tmp.push_back(all_values[i][j][l]);
					}
				val=vall_tmp;
				tab_tmp.push_back(val);
				vall_tmp.clear();
			}

		}
		all_values.clear();
		all_values.push_back(tab_tmp);
	return 0;
}
template <typename T>
void DataOperations<T>::setOutputFiles(string names)
{
  char delim = ',';
  string name;
  istringstream iss (names);
  while (getline(iss, name, delim))
  {
    file_out.push_back(name);
  }
}
